@@ -1,5 +1,6 @@
 import inspect, uuid, jsonpickle, random, string
-from .opio import Artifact
+
+from .opio import Artifact, BigParameter
 from ..op_template import PythonScriptOPTemplate
 from ..io import Inputs, Outputs, InputParameter, OutputParameter, InputArtifact, OutputArtifact, S3Artifact
 from ..utils import upload_artifact
@@ -11,7 +12,8 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                  output_artifact_archive=None, input_parameter_slices=None, output_artifact_slices=None,
                  output_parameter_slices=None, output_artifact_global_name=None, slices=None, python_packages=None,
                  timeout=None, retry_on_transient_error=None, output_parameter_default=None, output_parameter_global_name=None,
-                 timeout_as_transient_error=False, memoize_key=None, key=None, volumes=None, mounts=None):
+                 timeout_as_transient_error=False, memoize_key=None, volumes=None, mounts=None, image_pull_policy=None,
+                 cpu_requests=None, memory_requests=None, cpu_limits=None, memory_limits=None):
         """
         Convert from Python class OP to OP template
         :param op_class: Python class OP
@@ -32,9 +34,13 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         :param output_parameter_global_name: a dict specifying global names of output parameters within the workflow
         :param timeout_as_transient_error: regard timeout as transient error or fatal one
         :param memoize_key: memoized key of the OP template
-        :param key: key of the OP template
         :param volumes: volumes to use in the OP template
         :param mounts: volumes to mount in the OP template
+        :param image_pull_policy: Always, IfNotPresent, Never
+        :param cpu_requests: CPU requests
+        :param memory_requests: memory requests
+        :param cpu_limits: CPU limits
+        :param memory_limits: memory limits
         """
         class_name = op_class.__name__
         input_sign = op_class.get_input_sign()
@@ -47,9 +53,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 output_artifact_slices = {}
                 for name in slices.output_artifact:
                     output_artifact_slices[name] = slices.slices
-                    output_sign[name].save = S3Artifact(key=str(uuid.uuid4()) + "-{{workflow.duration}}") # stack slices to a S3Artifact for default
                     output_sign[name].archive = None # not archive for default
-                    output_sign[name].global_name = class_name + "-" + name # set global name for default
             if slices.output_parameter is not None: output_parameter_slices = {name: slices.slices for name in slices.output_parameter}
         if output_artifact_save is not None:
             for name, save in output_artifact_save.items():
@@ -61,7 +65,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
             for name, global_name in output_artifact_global_name.items():
                 output_sign[name].global_name = global_name
         super().__init__(name="%s-%s" % (class_name, "".join(random.sample(string.digits + string.ascii_lowercase, 5))), inputs=Inputs(), outputs=Outputs(),
-                volumes=volumes, mounts=mounts)
+                volumes=volumes, mounts=mounts, cpu_requests=cpu_requests, memory_requests=memory_requests, cpu_limits=cpu_limits, memory_limits=memory_limits)
         self.slices = slices
         if timeout is not None: self.timeout = "%ss" % timeout
         if retry_on_transient_error is not None:
@@ -74,12 +78,16 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         for name, sign in input_sign.items():
             if isinstance(sign, Artifact):
                 self.inputs.artifacts[name] = InputArtifact(path="/tmp/inputs/artifacts/" + name, optional=sign.optional)
+            elif isinstance(sign, BigParameter):
+                self.inputs.parameters[name] = InputParameter(save_as_artifact=True, path="/tmp/inputs/parameters/" + name)
             else:
                 self.inputs.parameters[name] = InputParameter()
         for name, sign in output_sign.items():
             if isinstance(sign, Artifact):
                 self.outputs.artifacts[name] = OutputArtifact(path="/tmp/outputs/artifacts/" + name,
                     archive=sign.archive, save=sign.save, global_name=sign.global_name)
+            elif isinstance(sign, BigParameter):
+                self.outputs.parameters[name] = OutputParameter(save_as_artifact=True, value_from_path="/tmp/outputs/parameters/" + name)
             else:
                 default = None
                 if output_parameter_default is not None and name in output_parameter_default:
@@ -108,7 +116,8 @@ class PythonOPTemplate(PythonScriptOPTemplate):
 
         if op_class.__module__ == "__main__":
             source_lines, start_line = inspect.getsourcelines(op_class)
-            pre_lines = open(inspect.getsourcefile(op_class), "r").readlines()[:start_line-1]
+            with open(inspect.getsourcefile(op_class), "r") as fd:
+                pre_lines = fd.readlines()[:start_line-1]
             script += "".join(pre_lines + source_lines) + "\n"
 
         script += "import os, sys, traceback, jsonpickle\n"
@@ -125,7 +134,10 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 script += "input['%s'] = handle_input_artifact('%s', input_sign['%s'], %s, '/tmp')\n" % (name, name, name, slices)
             else:
                 slices = self.get_slices(input_parameter_slices, name)
-                script += "input['%s'] = handle_input_parameter('%s', r'{{inputs.parameters.%s}}', input_sign['%s'], %s)\n" % (name, name, name, name, slices)
+                if isinstance(sign, BigParameter):
+                    script += "input['%s'] = handle_input_parameter('%s', '', input_sign['%s'], %s, '/tmp')\n" % (name, name, name, slices)
+                else:
+                    script += "input['%s'] = handle_input_parameter('%s', r'{{inputs.parameters.%s}}', input_sign['%s'], %s, '/tmp')\n" % (name, name, name, name, slices)
 
         script += "try:\n"
         script += "    output = op_obj.execute(input)\n"
@@ -150,6 +162,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 script += "handle_output_parameter('%s', output['%s'], output_sign['%s'], %s, '/tmp')\n" % (name, name, name, slices)
 
         self.image = image
+        self.image_pull_policy = image_pull_policy
         if command is not None:
             self.command = command
         else:
@@ -158,7 +171,6 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         self.init_progress = "%s/%s" % (op_class.progress_current, op_class.progress_total)
 
         self.memoize_key = memoize_key
-        self.key = key
 
     def get_slices(self, slices_dict, name):
         slices = None
