@@ -1,4 +1,5 @@
 from copy import deepcopy
+
 import jsonpickle
 from argo.workflows.client import (
     V1alpha1WorkflowStep,
@@ -7,7 +8,8 @@ from argo.workflows.client import (
     V1alpha1ContinueOn,
     V1alpha1ResourceTemplate
 )
-from .io import InputParameter, OutputParameter, PVC, ArgoVar
+from .common import S3Artifact
+from .io import InputArtifact, InputParameter, OutputArtifact, OutputParameter, PVC, ArgoVar
 from .op_template import ShellOPTemplate, PythonScriptOPTemplate
 from .utils import copy_file
 from .util_ops import CheckNumSuccess, CheckSuccessRatio
@@ -83,8 +85,7 @@ def if_expression(_if, _then, _else):
 
 class Step:
     def __init__(self, name, template, parameters=None, artifacts=None, when=None, with_param=None, continue_on_failed=False,
-            continue_on_num_success=None, continue_on_success_ratio=None, with_sequence=None, key=None, executor=None, use_resource=None,
-            use_template=None):
+            continue_on_num_success=None, continue_on_success_ratio=None, with_sequence=None, key=None, executor=None, use_resource=None):
         """
         Instantiate a step
         :param name: the name of the step
@@ -100,7 +101,6 @@ class Step:
         :param key: the key of the step
         :param executor: define the executor to execute the script
         :param use_resource: use k8s resource
-        :param use_template: use custom template
         :return:
         """
         self.name = name
@@ -114,6 +114,7 @@ class Step:
         self.continue_on_num_success = continue_on_num_success
         self.continue_on_success_ratio = continue_on_success_ratio
         self.check_step = None
+        self.prepare_step = None
 
         if parameters is not None:
             self.set_parameters(parameters)
@@ -127,18 +128,64 @@ class Step:
         self.key = key
         self.executor = executor
         self.use_resource = use_resource
-        self.use_template = use_template
+
+        if self.key is not None:
+            self.template.inputs.parameters["dflow_key"] = InputParameter(value="")
+            self.inputs.parameters["dflow_key"] = InputParameter(value=str(self.key))
+
+        if hasattr(self.template, "slices") and self.template.slices is not None and self.template.slices.output_artifact is not None:
+            new_template = deepcopy(self.template)
+            new_template.name = self.template.name + "-" + self.name
+            script = ""
+            for name in new_template.slices.output_artifact:
+                script += "mkdir -p /tmp/outputs/artifacts/%s\n" % name
+                script += "echo '{\"path_list\": []}' > /tmp/outputs/artifacts/%s/.dflow.init\n" % name
+            init_template = ShellOPTemplate(name="%s-init-artifact" % new_template.name, image=new_template.image,
+                image_pull_policy=new_template.image_pull_policy, script=script)
+            if self.key is not None:
+                new_template.inputs.parameters["dflow_group_key"] = InputParameter(value="")
+                self.inputs.parameters["dflow_group_key"] = InputParameter(value=str(self.key).replace("{{item}}", "group"))
+                init_template.inputs.parameters["dflow_group_key"] = InputParameter()
+                for name in new_template.slices.output_artifact:
+                    init_template.outputs.artifacts[name] = OutputArtifact(path="/tmp/outputs/artifacts/%s" % name,
+                        save=S3Artifact(key="{{workflow.name}}/{{inputs.parameters.dflow_group_key}}/%s" % name), archive=None)
+                    new_template.outputs.artifacts[name].save.append(S3Artifact(
+                        key="{{workflow.name}}/{{inputs.parameters.dflow_group_key}}/%s" % name))
+                self.prepare_step = Step(name="%s-init-artifact" % self.name, template=init_template,
+                    parameters={"dflow_group_key": str(self.key).replace("{{item}}", "group")})
+            else:
+                new_template.inputs.parameters["dflow_artifact_key"] = InputParameter(value="")
+                self.inputs.parameters["dflow_artifact_key"] = InputParameter(value="{{workflow.name}}/{{steps.%s-init-artifact.id}}" % self.name)
+                for name in new_template.slices.output_artifact:
+                    init_template.outputs.artifacts[name] = OutputArtifact(path="/tmp/outputs/artifacts/%s" % name,
+                        save=S3Artifact(key="{{workflow.name}}/{{pod.name}}/%s" % name), archive=None)
+                    new_template.outputs.artifacts[name].save.append(S3Artifact(
+                        key="{{inputs.parameters.dflow_artifact_key}}/%s" % name))
+                self.prepare_step = Step(name="%s-init-artifact" % self.name, template=init_template)
+            for name in new_template.slices.output_artifact:
+                self.outputs.artifacts[name].redirect = self.prepare_step.outputs.artifacts[name]
+            self.template = new_template
 
     def __repr__(self):
         return self.id
     
     def set_parameters(self, parameters):
         for k, v in parameters.items():
+            # if a parameter is saved as artifact, the parameters it pass value to or its value comes from must be saved as artifact as well
+            if self.inputs.parameters[k].save_as_artifact and isinstance(v, (InputParameter, OutputParameter)):
+                v.save_as_artifact = True
+            if isinstance(v, (InputParameter, OutputParameter)) and v.save_as_artifact:
+                self.inputs.parameters[k].save_as_artifact = True
+
+            if self.inputs.parameters[k].save_as_artifact and isinstance(v, (InputParameter, OutputParameter, InputArtifact, OutputArtifact)):
+                self.inputs.parameters[k].source = v
+                return
+
             if v is None:
                 self.inputs.parameters[k].value = "null"
             else:
                 self.inputs.parameters[k].value = v
-    
+
     def set_artifacts(self, artifacts):
         for k, v in artifacts.items():
             if v is None:
@@ -147,19 +194,17 @@ class Step:
             else:
                 self.inputs.artifacts[k].source = v
 
-    def convert_to_argo(self):
-        if self.key is not None:
-            self.template.key = self.key
-            self.inputs.parameters["dflow_key"] = InputParameter(value=str(self.key))
-            if hasattr(self.template, "slices") and self.template.slices is not None and self.template.slices.output_artifact is not None:
-                self.inputs.parameters["dflow_group_key"] = InputParameter(value=str(self.key).replace("{{item}}", "group"))
+    def convert_to_argo(self, context=None):
         argo_parameters = []
+        argo_artifacts = []
         for par in self.inputs.parameters.values():
-            argo_parameters.append(par.convert_to_argo())
+            if par.save_as_artifact:
+                argo_artifacts.append(par.convert_to_argo())
+            else:
+                argo_parameters.append(par.convert_to_argo())
 
         new_template = None
 
-        argo_artifacts = []
         pvc_arts = []
         for art in self.inputs.artifacts.values():
             if isinstance(art.source, PVC):
@@ -226,15 +271,13 @@ class Step:
             elif (isinstance(new_template, PythonScriptOPTemplate)):
                 new_template.outputs.parameters["dflow_success_tag"] = OutputParameter(value_from_path="/tmp/success_tag", default="0")
                 new_template.script += "\n"
-                new_template.script += "open('/tmp/success_tag', 'w').write('1')\n"
+                new_template.script += "with open('/tmp/success_tag', 'w') as f:\n    f.write('1')\n"
             else:
                 raise RuntimeError("Unsupported type of OPTemplate for continue_on_num_success or continue_on_success_ratio")
 
         if new_template is not None:
             self.template = new_template
-            self.inputs = deepcopy(self.template.inputs)
             self.outputs = deepcopy(self.template.outputs)
-            self.inputs.set_step_id(self.id)
             self.outputs.set_step_id(self.id)
 
         if self.continue_on_num_success is not None:
@@ -257,23 +300,17 @@ class Step:
         if isinstance(self.with_param, ArgoVar):
             self.with_param = "{{=%s}}" % self.with_param.expr
 
+        if context is not None:
+            self.template = context.render(self.template)
+
         if self.executor is not None:
-            new_template = ShellOPTemplate(name=self.template.name + "-remote", inputs=self.template.inputs,
-                outputs=self.template.outputs, image=self.executor.image, command=self.executor.command, script=None, volumes=self.template.volumes, mounts=self.template.mounts,
-                init_progress=self.template.init_progress, timeout=self.template.timeout, retry_strategy=self.template.retry_strategy,
-                memoize_key=self.template.memoize_key, key=self.template.key)
-            new_template.script = self.executor.get_script(self.template.command, self.template.script)
-            self.template = new_template
+            self.template = self.executor.render(self.template)
 
         if self.use_resource is not None:
             self.template.resource = V1alpha1ResourceTemplate(action=self.use_resource.action,
                 success_condition=self.use_resource.success_condition, failure_condition=self.use_resource.failure_condition,
                 manifest=self.use_resource.get_manifest(self.template.command, self.template.script))
 
-        if self.use_template is not None:
-            self.use_template.render(self.template)
-            self.template = self.use_template
-
         return V1alpha1WorkflowStep(
             name=self.name, template=self.template.name, arguments=V1alpha1Arguments(
                 parameters=argo_parameters,
