@@ -1,11 +1,21 @@
 import os
 import uuid
 import shutil
+import string
+import random
 import tarfile
 import jsonpickle
 from minio import Minio
 from minio.api import CopySource
-from .io import S3Artifact, randstr
+from .common import S3Artifact
+
+s3_config = {
+    "endpoint": "127.0.0.1:9000",
+    "access_key": "admin",
+    "secret_key": "password",
+    "secure": False,
+    "bucket_name": "my-bucket"
+}
 
 def download_artifact(artifact, extract=True, **kwargs):
     """
@@ -42,6 +52,7 @@ def download_artifact(artifact, extract=True, **kwargs):
                     merge_dir(tmpdir, path)
                 shutil.rmtree(tmpdir)
 
+        remove_empty_dir_tag(path)
         return assemble_path_list(path, remove=True)
     else:
         raise NotImplementedError()
@@ -118,8 +129,18 @@ def copy_artifact(src, dst):
     copy_s3(src_key, dst_key)
     return S3Artifact(key=dst_key)
 
-def download_s3(key, path=None, recursive=True, endpoint="127.0.0.1:9000",
-            access_key="admin", secret_key="password", secure=False, bucket_name="my-bucket", **kwargs):
+def download_s3(key, path=None, recursive=True, endpoint=None, access_key=None, secret_key=None,
+        secure=None, bucket_name=None, **kwargs):
+    if endpoint is None:
+        endpoint = s3_config["endpoint"]
+    if access_key is None:
+        access_key = s3_config["access_key"]
+    if secret_key is None:
+        secret_key = s3_config["secret_key"]
+    if secure is None:
+        secure = s3_config["secure"]
+    if bucket_name is None:
+        bucket_name = s3_config["bucket_name"]
     if path is None:
         path = "."
     client = Minio(endpoint=endpoint, access_key=access_key, secret_key=secret_key, secure=secure)
@@ -137,8 +158,18 @@ def download_s3(key, path=None, recursive=True, endpoint="127.0.0.1:9000",
         client.fget_object(bucket_name=bucket_name, object_name=key, file_path=path)
     return path
 
-def upload_s3(path, key=None, endpoint="127.0.0.1:9000", access_key="admin", secret_key="password",
-            secure=False, bucket_name="my-bucket", **kwargs):
+def upload_s3(path, key=None, endpoint=None, access_key=None, secret_key=None, secure=None,
+        bucket_name=None, **kwargs):
+    if endpoint is None:
+        endpoint = s3_config["endpoint"]
+    if access_key is None:
+        access_key = s3_config["access_key"]
+    if secret_key is None:
+        secret_key = s3_config["secret_key"]
+    if secure is None:
+        secure = s3_config["secure"]
+    if bucket_name is None:
+        bucket_name = s3_config["bucket_name"]
     if key is None:
         key = "upload/%s/%s" % (uuid.uuid4(), os.path.basename(path))
     client = Minio(endpoint=endpoint, access_key=access_key, secret_key=secret_key, secure=secure)
@@ -155,8 +186,18 @@ def upload_s3(path, key=None, endpoint="127.0.0.1:9000", access_key="admin", sec
                 client.fput_object(bucket_name=bucket_name, object_name="%s%s/%s" % (key, rel_path, f), file_path=os.path.join(dn, f))
     return key
 
-def copy_s3(src_key, dst_key, recursive=True, endpoint="127.0.0.1:9000", access_key="admin", secret_key="password",
-            secure=False, bucket_name="my-bucket", **kwargs):
+def copy_s3(src_key, dst_key, recursive=True, key=None, endpoint=None, access_key=None, secret_key=None,
+        secure=None, bucket_name=None, **kwargs):
+    if endpoint is None:
+        endpoint = s3_config["endpoint"]
+    if access_key is None:
+        access_key = s3_config["access_key"]
+    if secret_key is None:
+        secret_key = s3_config["secret_key"]
+    if secure is None:
+        secure = s3_config["secure"]
+    if bucket_name is None:
+        bucket_name = s3_config["bucket_name"]
     client = Minio(endpoint=endpoint, access_key=access_key, secret_key=secret_key, secure=secure)
     if recursive:
         for obj in client.list_objects(bucket_name=bucket_name, prefix=src_key, recursive=True):
@@ -190,13 +231,14 @@ def copy_file(src, dst, func=os.link):
         raise RuntimeError("File %s not found" % src)
 
 def assemble_path_list(art_path, remove=False):
-    path_list = [art_path]
+    path_list = []
     if os.path.isdir(art_path):
         dflow_list = []
         for f in os.listdir(art_path):
             if f[:6] == ".dflow":
-                for item in jsonpickle.loads(open('%s/%s' % (art_path, f), 'r').read())['path_list']:
-                    if item not in dflow_list: dflow_list.append(item) # remove duplicate
+                with open('%s/%s' % (art_path, f), 'r') as fd:
+                    for item in jsonpickle.loads(fd.read())['path_list']:
+                        if item not in dflow_list: dflow_list.append(item) # remove duplicate
                 if remove:
                     os.remove(os.path.join(art_path, f))
         if len(dflow_list) > 0:
@@ -206,3 +248,11 @@ def assemble_path_list(art_path, remove=False):
 def convert_dflow_list(dflow_list):
     dflow_list.sort(key=lambda x: x['order'])
     return list(map(lambda x: x['dflow_list_item'], dflow_list))
+
+def remove_empty_dir_tag(path):
+    for dn, ds, fs in os.walk(path, followlinks=True):
+        if ".empty_dir" in fs:
+            os.remove(os.path.join(dn, ".empty_dir"))
+
+def randstr(l=5):
+    return "".join(random.sample(string.digits + string.ascii_lowercase, l))
